package net.osmand.plus.views;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Serializable;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.http.entity.SerializableEntity;

import net.osmand.LogUtil;
import net.osmand.data.ESBRData;
import net.osmand.osm.LatLon;
import net.osmand.plus.R;
import net.osmand.plus.geoassist.GeoHelper;
import net.osmand.plus.routing.RouteProvider;
import net.osmand.plus.views.GeoPluginLayer.GeoPlugin;
import alice.tuprolog.Int;
import alice.util.IPAddress;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.PointF;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Paint.Cap;
import android.graphics.Paint.Join;
import android.graphics.Paint.Style;
import android.graphics.drawable.Drawable;
import android.location.Location;
import android.net.wifi.WifiManager;
import android.os.Environment;
import android.os.Handler;
import android.os.Looper;
import android.telephony.TelephonyManager;
import android.text.format.DateFormat;
import android.widget.Toast;

public class ESBRPluginLayer extends GeoPluginLayer {
	private static final org.apache.commons.logging.Log log = LogUtil.getLog(ESBRPluginLayer.class);
	private static final boolean LOGGING=true;
	private static final float MIN_DISTANCE=100;
	private static final float JOINT_DISTANCE= 15;
	private static final int ESBR_PORT=50000;
	private static final int ALERT_TIMEOUT=8000; //Milliseconds
	private static final int ALERT_INTERVAL=500;
	private static final int MAX_ROUTE_DISTANCE=1000; //Meters
	private static final int MAX_ROUTE_POINTS=45;
	private static final int MAX_PACKET_SIZE=1500;
	private static final int MAX_CACHE_MSG = 1000;
	
	private DatagramSocket generatorSocket;
	private DatagramSocket protocolSocket;
	
	private OsmandMapTileView view;
	private Bitmap emergencyIcon;
	private Bitmap otherIcon;
	private Drawable panicButtonIcon;
	private Drawable routingButtonIcon;
	private Paint paint;
	
	private PrintWriter pw;
	
	ESBRProtocol protocol;
	ESBRPacketGenerator generator;
	
	private Handler uiHandler;
	
	@Override
	public GeoPlugin GetType() {
		return GeoPlugin.EMDR;
	}

	public ESBRPluginLayer(GeoHelper geoHelper) {
		super(geoHelper);
	}
	
	@Override
	public void initLayer(OsmandMapTileView view) {
		long nodeId = 0;
		TelephonyManager info = (TelephonyManager)view.getApplication().getSystemService(Context.TELEPHONY_SERVICE);
		String nodeImei =  info.getDeviceId();
		
		uiHandler = new Handler();
		
		Date now = new Date(System.currentTimeMillis());
		java.text.DateFormat format = new SimpleDateFormat("d_kkmm");
		String date = format.format(now);
		String filename= Environment.getExternalStorageDirectory()+"/ESBR"+date+"_"+nodeImei+".log";
		try{
			pw = new PrintWriter(filename);
		}catch (FileNotFoundException e) {
			String msg= new String("Log can't be activated");
			showMessage(msg , msg.length());
		}
		
		try {
			nodeId = Long.parseLong(nodeImei);
			if(protocolSocket == null)
			{
				protocolSocket=new DatagramSocket(ESBR_PORT);
			}
			else if(protocolSocket.isClosed())
			{
				protocolSocket.bind(new InetSocketAddress(ESBR_PORT));
			}
			protocolSocket.setBroadcast(true);
		} catch (SocketException e1) {
			showMessage(e1.toString(), e1.toString().length());
		}
		
		
		protocol= new ESBRProtocol( nodeId, protocolSocket);
		protocol.start();
		
		try{
			if(generatorSocket==null){
				generatorSocket= new DatagramSocket(ESBR_PORT-1);
				
			}
			else if(generatorSocket.isClosed()){
				generatorSocket.bind(new InetSocketAddress(ESBR_PORT-1));
			}
			generatorSocket.setBroadcast(true);
			generator = new ESBRPacketGenerator(nodeId, generatorSocket);
			generator.start();
		}
		catch(SocketException e){
			System.out.println(e.getMessage());
		}
 		
		paint = new Paint();
		paint.setColor(view.getResources().getColor(R.color.nav_track_fluorescent));
		paint.setAntiAlias(true);
		paint.setFilterBitmap(true);
		paint.setDither(true);
		paint.setStyle(Style.STROKE);
		paint.setStrokeWidth(14);
		paint.setStrokeCap(Cap.ROUND);
		paint.setStrokeJoin(Join.ROUND);
		this.view= view;
		emergencyIcon = BitmapFactory.decodeResource(view.getResources(), R.drawable.emergency_icon);
		otherIcon = BitmapFactory.decodeResource(view.getResources(), R.drawable.other_location);
	}

	@Override
	public void onDraw(Canvas canvas, RectF latLonBounds, RectF tilesRect,
			DrawSettings settings) {
		//Print the panic button to activate packet Generator.
		Rect bounds=canvas.getClipBounds();
		//Place the button in the middle of right side
		if(generator.inAlarm){
			panicButtonIcon = view.getResources().getDrawable(R.drawable.panic_button_pressed);
		}
		else{
			panicButtonIcon = view.getResources().getDrawable(R.drawable.panic_button);
		}
		panicButtonIcon.setBounds(bounds.right-panicButtonIcon.getIntrinsicWidth(), bounds.top+(bounds.bottom-panicButtonIcon.getIntrinsicHeight())/2, bounds.right, bounds.top+(bounds.bottom+panicButtonIcon.getIntrinsicHeight())/2);
		panicButtonIcon.draw(canvas);
		
		//Place the routing button in the middle of left side
		if(protocol.isResendActive() && protocol.isResendAll()){
			routingButtonIcon = view.getResources().getDrawable(R.drawable.green_router);
		}
		else if(protocol.isResendActive() && !protocol.isResendAll()){
			routingButtonIcon = view.getResources().getDrawable(R.drawable.yellow_router);
		}
		else{
			routingButtonIcon = view.getResources().getDrawable(R.drawable.red_router);
		}
		routingButtonIcon.setBounds(bounds.left, bounds.top+(bounds.bottom-routingButtonIcon.getIntrinsicHeight())/2, bounds.left+routingButtonIcon.getIntrinsicWidth(), bounds.top+(bounds.bottom+routingButtonIcon.getIntrinsicHeight())/2);
		routingButtonIcon.draw(canvas);
		
		//Print an ambulance if alert
		Map<Long, Boolean> warningList = protocol.getWarnings();
		Map<Long, Location> alertList = protocol.getAlertLocations();
		Map<Long, LatLon[]> routes = protocol.getRoutes();
		for(Long nodeId: alertList.keySet())
		{
			if(warningList.containsKey(nodeId))
			{
				if(routes.get(nodeId).length > 1)
				{
					Path path= new Path();
					LatLon nodeO= routes.get(nodeId)[0];

					int x = view.getRotatedMapXForPoint(nodeO.getLatitude(), nodeO.getLongitude());
					int y = view.getRotatedMapYForPoint(nodeO.getLatitude(), nodeO.getLongitude());

					path.moveTo(x, y);

					LatLon[] route=routes.get(nodeId);
					for(LatLon node: route)
					{
						x = view.getRotatedMapXForPoint(node.getLatitude(), node.getLongitude());
						y = view.getRotatedMapYForPoint(node.getLatitude(), node.getLongitude());					
						path.lineTo(x, y);
					}
					canvas.drawPath(path, paint);
				}

				Location o = alertList.get(nodeId);
				if (o.getLatitude() >= latLonBounds.bottom && o.getLatitude() <= latLonBounds.top  && o.getLongitude() >= latLonBounds.left 
						&& o.getLongitude() <= latLonBounds.right ) {
					int x = view.getRotatedMapXForPoint(o.getLatitude(), o.getLongitude());
					int y = view.getRotatedMapYForPoint(o.getLatitude(), o.getLongitude());
					canvas.drawBitmap(emergencyIcon, x - emergencyIcon.getWidth() / 2, 
							y - emergencyIcon.getHeight() / 2, paint);
				}
			}
			else
			{
				Location o = alertList.get(nodeId);
				if (o.getLatitude() >= latLonBounds.bottom && o.getLatitude() <= latLonBounds.top  && o.getLongitude() >= latLonBounds.left 
						&& o.getLongitude() <= latLonBounds.right ) {
					int x = view.getRotatedMapXForPoint(o.getLatitude(), o.getLongitude());
					int y = view.getRotatedMapYForPoint(o.getLatitude(), o.getLongitude());
					canvas.drawBitmap(otherIcon, x - otherIcon.getWidth() / 2, 
							y - otherIcon.getHeight() / 2, paint);
				}
			}
		}

	}

	@Override
	public void destroyLayer() {
		pw.close();
		protocolSocket.close();
		generatorSocket.close();
		generator.interrupt();
		protocol.interrupt();
	}

	@Override
	public boolean onSingleTap(PointF point) {
		Rect bounds=panicButtonIcon.getBounds();
		if(bounds.contains((int)point.x, (int)point.y))
		{
			onPanicButtonPress();
			return true;
		}
		
		bounds = routingButtonIcon.getBounds();
		if(bounds.contains((int)point.x, (int)point.y))
		{
			onRoutingButtonPress();
			return true;
		}
		
		return false;
	}

	private synchronized void logMessage(String msg, String name){
		if(pw!=null && LOGGING){
			pw.println(System.currentTimeMillis()+"\t"+name+"\t"+msg);
			pw.flush();
		}
	}
	
	private void onPanicButtonPress() {
		generator.changeState();
		view.refreshMap(false);
	}

	private void onRoutingButtonPress() {
		if(protocol.isResendActive() && !protocol.isResendAll()){
			protocol.setResendAll(true);
		}
		else if (protocol.isResendActive() && protocol.isResendAll()){
			protocol.setResendActive(false);
			protocol.setResendAll(false);
		}
		else{
			protocol.setResendActive(true);
		}
		view.refreshMap(false);
	}
	
	private void showMessage(final String msg, final int length) {
		uiHandler.post(new Runnable() {
			@Override
			public void run() {
				Toast.makeText(view.getContext(), msg, Toast.LENGTH_SHORT).show();
			}
		});
	}
	
	private class ESBRProtocol extends Thread{
		volatile Map<Long, Location> alertLocations= new HashMap<Long, Location>();
		volatile Map<Long, LatLon[]> routes = new HashMap<Long, LatLon[]>();
		volatile Map<Long, Thread> monitorList= new HashMap<Long, Thread>();
		volatile Map<Long, Boolean> warnings = new HashMap<Long, Boolean>();
		volatile Map<Long, Long> streets=new HashMap<Long, Long>();
		SortedSet<ESBRData> received = new TreeSet<ESBRData>();
		volatile long nodeId;
		
		private DatagramSocket socket;
		
		private class TimerOut extends Thread{
			private Long id;
			
			TimerOut(Long id){
				this.id=id;
			}
			
			public void run(){
				while(true)
				{
					try {
						sleep(ALERT_TIMEOUT,0);
						logMessage("Timeout for alert "+id, this.getClass().toString());
						removeFromDb(id);
						view.refreshMap(false);
						return;
					} catch (InterruptedException e) 
					{
					} catch (Exception e){
						String msg= new String( "Timer"+e.toString());
						showMessage(msg, msg.length());
					}
				}
			}
		};
		
		private volatile boolean resendAll;
		private volatile boolean resendActive;
		
		public ESBRProtocol(long nodeID, DatagramSocket socket) {
			this.nodeId=nodeID;
			this.socket=socket;
			this.resendAll=false;
			this.resendActive=true;
		}
		
		
		public Map<Long, Location> getAlertLocations() {
			return alertLocations;
		}
		
		public Map<Long, LatLon[]> getRoutes() {
			return routes;
		}
		
		public Map<Long, Boolean> getWarnings() {
			return warnings;
		}


		public boolean isInAlert() {
			return !warnings.isEmpty();
		}

		public boolean isResendAll() {
			return resendAll;
		}
		
		public void setResendAll(boolean resendAll) {
			this.resendAll = resendAll;
		}
		
		public boolean isResendActive() {
			return resendActive;
		}
		
		public void setResendActive(boolean resendActive) {
			this.resendActive = resendActive;
		}

		@Override
		public void run() {
			while(true)
			{
				DatagramPacket packet = new DatagramPacket(new byte[MAX_PACKET_SIZE], MAX_PACKET_SIZE);
				try{
					socket.receive(packet);

					ByteArrayInputStream bis = new ByteArrayInputStream(packet.getData());
					ObjectInputStream is = new ObjectInputStream(bis);
					ESBRData esbrData = (ESBRData)is.readObject();
					boolean result;
					if(esbrData.getNodeID()!= nodeId){
						result = received.add(esbrData);
					}
					else{
						result=false;
					}
					
					if(received.size() >= MAX_CACHE_MSG){
						ESBRData remove=received.first();
						received.remove(remove);
					}
					
					if(result)
					{
						Location myLoc = geoHelper.getLocOnStreet();
						if(myLoc!=null)
						{
							Location to = new Location("ESBR_PROTOCOL");
							to.setLatitude(esbrData.getRoute()[0].getLatitude());
							to.setLongitude(esbrData.getRoute()[0].getLongitude());
							logMessage("MyLoc "+myLoc.getLatitude()+" "+myLoc.getLongitude()+ "Rx \t nodeId "+esbrData.getNodeID()+"\t MsgId "+esbrData.getMsgID()+" Location "+esbrData.getRoute()[0].getLatitude()+" "+esbrData.getRoute()[0].getLongitude()+"\t Warning "+esbrData.isWarning()+"\tDistance "+ myLoc.distanceTo(to) , "PacketReceived");
						}
						Location rxLoc= new Location("ESBR");
						rxLoc.setLatitude(esbrData.getRoute()[0].getLatitude());
						rxLoc.setLongitude(esbrData.getRoute()[0].getLongitude());

						addToDb(esbrData.getNodeID(), rxLoc, esbrData.isWarning(), esbrData.getStreetID(), esbrData.getRoute());

						view.refreshMap(false);

						if(esbrData.isWarning() && (resend(esbrData) || resendAll ) && resendActive )
						{
							logMessage("ResendTx\tMsgId "+esbrData.getMsgID()+"\tnodeId "+esbrData.getNodeID()+" MyStreet "+geoHelper.getStreetID()+" ItsStreet "+esbrData.getStreetID(), "PacketRetransmitted");
							ByteArrayOutputStream bos=new ByteArrayOutputStream();
							ObjectOutputStream os=new ObjectOutputStream(bos);
							os.writeObject(esbrData);

							packet = new DatagramPacket(bos.toByteArray(), 
									bos.toByteArray().length, InetAddress.getByName("255.255.255.255"), ESBR_PORT);
							//String msg= new String("Mensaje Reenviado");
							//showMessage(msg, msg.length());
							socket.send(packet);
						}
					}
				}
				catch(SocketException e){
					if(e!=null){
						if(protocolSocket.isClosed()){
							return;
						}
						String msg= new String( "Protocol"+e.toString());
						showMessage(msg, msg.length());
					}
				}
				catch (Exception e){
					String msg= new String( "Protocol"+e.toString());
					showMessage(msg, msg.length());
					return;
				}
			}
		}
		
		private boolean resend(ESBRData data){
			Location rxLoc= new Location("ESBR");
			Location myLoc = geoHelper.getLocOnStreet();
			if(myLoc == null)
			{
				return false;
			}
			rxLoc.setLatitude(data.getRoute()[0].getLatitude());
			rxLoc.setLongitude(data.getRoute()[0].getLongitude());
			float distance= myLoc.distanceTo(rxLoc);
			long myStreet = geoHelper.getStreetID();
			if(distance > MIN_DISTANCE || myStreet != data.getStreetID() || isInJoint() ){
				logMessage("Dist "+distance+" MyStreet "+geoHelper.getStreetID()+" ItsStreet "+data.getStreetID()+"\tIsJoint "+isInJoint(), "Resend" );
				return true;
			}
			return false;
		}
		
		private boolean isInJoint(){
			for (Long nodeId : streets.keySet()) {
				if(streets.get(nodeId)!=geoHelper.getStreetID()){
					return true;
				}
			}
			if(geoHelper.getDistanceToClosestIntersection()< JOINT_DISTANCE){
				return true;
			}
			
			return false;
		}
		
		private synchronized void addToDb(long nodeId, Location loc, boolean warning, long streetId, LatLon[] route){
			Thread monitor = monitorList.get(nodeId);
			
			if( monitor != null )
			{
				if(monitor.isAlive())
				{
					monitor.interrupt();
				}
				else
				{
					monitor.start();
				}
			}
			else
			{
				monitor=new TimerOut(nodeId);
				monitorList.put(nodeId, monitor);
				monitor.start();
			}
			
			alertLocations.put(nodeId, loc);
			if(warning){
				warnings.put(nodeId, warning);
			}
			else{
				warnings.remove(nodeId);
			}
			streets.put(nodeId, streetId);
			routes.put(nodeId,route);
		}
		
		private synchronized void removeFromDb(long id){
			alertLocations.remove(id);
			monitorList.remove(id);
			routes.remove(id);
			warnings.remove(id);
			streets.remove(id);
		}
	}
		
	private class ESBRPacketGenerator extends Thread{
		private final int BURST_SIZE=200;
		private int msgID = 0;
		private int counter;
		private long nodeID ;
		private DatagramSocket socket;
		
		private boolean inAlarm=false; 
	
		ESBRPacketGenerator(long nodeID, DatagramSocket socket) {
			this.nodeID = nodeID;
			this.socket = socket;
		}
		
		public synchronized void changeState(){
			String msg;
			if(inAlarm == false){
				msg = new String("Alerta Iniciada");
				counter=BURST_SIZE;
			}
			else{
				msg = new String("Alerta detenida");
				counter=0;
			}
			msg=msg+"\n"+msgID;
			inAlarm = !inAlarm;
			showMessage(msg, msg.length());
		}
		
		public void run(){
			while(true){
				try {
					sleep(ALERT_INTERVAL, 0);
					
					Location loc;
					
					/*if(geoHelper.isOnStreet()){
						loc= geoHelper.getLocOnStreet();
					}
					else{*/
						loc= geoHelper.getLocOnStreet();
					//}
					if(loc!=null)
					{
						List<Location> route = geoHelper.getCurrentRoute();
						List<LatLon> outRoute=new LinkedList<LatLon>();
						outRoute.add(new LatLon(loc.getLatitude(),loc.getLongitude()));
						if(route!=null && inAlarm == true)
						{
							double dist=loc.distanceTo(route.get(0));
							for(int i=1; i<route.size() && i< MAX_ROUTE_POINTS && dist < MAX_ROUTE_DISTANCE;i++)
							{
								dist+=route.get(i).distanceTo(route.get(i-1));
								outRoute.add(new LatLon(route.get(i-1).getLatitude(),route.get(i-1).getLongitude()));
							}
						}

						ESBRData data = new ESBRData(nodeID, msgID++, geoHelper.getStreetID() , inAlarm, outRoute.toArray(new LatLon[0]));
						if(inAlarm){
							logMessage("Node "+nodeID+"\tmsgId "+msgID+ " Location "+ loc.getLatitude() +" "+ loc.getLongitude(), "PacketSent");
							if(--counter == 0){
								changeState();
							}
						}
						ByteArrayOutputStream bos=new ByteArrayOutputStream();
						ObjectOutputStream os=new ObjectOutputStream(bos);
						os.writeObject(data);

						DatagramPacket packet = new DatagramPacket(bos.toByteArray(), 
								bos.toByteArray().length, InetAddress.getByName("255.255.255.255"), ESBR_PORT);

						socket.send(packet);
					}
				}
				catch(SocketException e){
					String msg= new String( "Generator "+e.toString());
					if(socket.isClosed())
					{
						return;
					}
					
					showMessage(msg, msg.length());
					if(inAlarm){
						changeState();
					}
				}
				catch(InterruptedException e)
				{
					return;
				}
				catch (Exception e){
					String msg= new String( "Generator "+e.toString());
					showMessage(msg, msg.length());
					return;
				}
			}
		}
		
	}
}
